#  Copyright 2021 Technische Universiteit Delft
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import numpy as np

# Create VTT midpoints
def vtt_midpoints(vtt_grid):
    
    # Calculate midpoints
    vtt_grid_midpoints = ((np.append(vtt_grid, 0) + np.append(0,vtt_grid)))/2
    
    # Replace last value of the VTT midpoints by the last value of the VTT array
    vtt_grid_midpoints[-1] = vtt_grid[-1]

    return vtt_grid_midpoints

# Create predicted VTT for each respondent, based in the choice probability and a VTT grid.
def predicted_vtt(p,vtt_grid,NP):

    # Obtain probabilities from the CDF plot
    pp = ((np.append(p, 0)-np.append(0,p)))
    pp[pp<0] = 0
    pp = (np.divide(pp,np.sum(pp)))
    pp[-1] = 0
    pp = np.append(0,pp)

    pp = ((np.append(p, 0)-np.append(0,p)))
    pp[-1] = 0
    pp = np.append(0,pp)

    count_data = []
    for n in range(0,len(vtt_grid)):
        dat = [vtt_grid[n]]*np.round((pp[n]*NP)).astype(int)
        count_data = np.append(count_data,dat)

    return count_data

# from random import uniform

# def predicted_vtt(ecdf,grid,NP):
#     pp = ((np.append(ecdf, 0)-np.append(0,ecdf)))
#     count_data = []
#     for n in range(1,len(grid)):
#         dat = [uniform(grid[n-1], grid[n]) for p in range(0, np.round((pp[n]*NP)).astype(int))]
#         count_data = np.append(count_data,dat)
#     return count_data


#  Copyright 2021 Technische Universiteit Delft
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""Modules to configure and estimate a Logistic regression-based model."""
from dataclasses import dataclass
from typing import Optional, Tuple
from scipy.optimize import minimize
from numdifftools import Hessian
import numpy as np
import warnings

from py_np4vtt.data_format import ModelArrays

warnings.filterwarnings('ignore')

@dataclass
class ConfigLogistic:
    """Configuration class of the logistic regression (ANN-0) model.
    
    This class stores the configuration parameters of a logistic regression 
    model and performs integrity checks before being passed to the model 
    object.
    
    Parameters
    ----------
    startScale : float
        The starting value of the scale parameter.
    startIntercept : float
        The starting value of the intercept parameter.
    startParameter: float
        The starting value of the VTT parameter.
    maxIterations: int
        Maximum number of iterations of the estimation routine.
    seed: Optional[int]
        Random seed
    """
    startScale: float
    startIntercept: float
    startParameter: float

    maxIterations: int

    seed: Optional[int]

    def validate(self):
        # Create errormessage list
        errorList = []

        if not self.startScale > 0:
            errorList.append('Scale starting value must be positive.')

        if not self.maxIterations > 0:
            errorList.append('Max iterations must be greater than zero.')

        if not self.seed >= 0:
            errorList.append('Seed must be non-negative.')

        # Whoever calls this validator knows that empty errorList means validator success
        return errorList

class ModelLogistic:
    """Logistic regression model.
    
    This is the model class that prepares the data and estimates 
    the logistic regression.
    
    Parameters
    -----------
    params : ConfigLogistic
        A configuration class of a logistic regression model.
    arrays : ModelArrays
        Model arrays created with `make_modelarrays`
        
    Attributes
    ----------
    None.
    """
    def __init__(self, cfg: ConfigLogistic, arrays: ModelArrays):
        self.cfg = cfg
        self.arrays = arrays

    def run(self):
        """Estimates the logistic regression model.
        
        Parameters
        ----------
        None.

        Returns
        -------
        x : np.ndarray
            The estimated parameters of the scale, intercept and VTT parameter.
        se : np.ndarray
            The standard errors of the estimated parameters
        vtt: np.ndarray
            The estimated VTT per respondent, based in the estimated 
            parameters and the sample.
        init_ll : float
            Log-likelihood at the starting values
        ll : float
            Log-likelihood in the optimum.
        exitflag : int
            Exit flag of the optimisation routine. If `exitflag=0`, the 
            optimisation succeeded. Otherwise, check the configuration parameters.
        """
        # Use passed seed if desired
        if self.cfg.seed:
            np.random.seed(self.cfg.seed)

        # Prepare data
        i_obs_y = \
            np.tile(np.random.randint(1, self.arrays.T, size=(self.arrays.NP, 1)), (1, self.arrays.T)) == \
            np.tile(np.arange(1, self.arrays.T + 1), (self.arrays.NP, 1))
        
        i_obs_x = (i_obs_y == 0)

        # Set vector of starting values of parameters to estimate
        x0 = np.array([self.cfg.startScale, self.cfg.startIntercept, self.cfg.startParameter])

        BVTT=np.sum(i_obs_y * self.arrays.BVTT, axis=1)
        sumYBVTT=np.sum(i_obs_x * self.arrays.BVTT * self.arrays.Choice, axis=1)
        y_regress=np.sum(self.arrays.Choice * i_obs_y, axis=1)

        # LL at the start values
        init_ll = -ModelLogistic.objectiveFunction(x0, sumYBVTT, BVTT, y_regress)

        # Starting arguments and values for minimizer
        argTuple = (sumYBVTT, BVTT, y_regress)
        
        # Start minimization routine
        results = minimize(ModelLogistic.objectiveFunction, x0, args=argTuple, method='L-BFGS-B',options={'gtol': 1e-6,'maxiter': self.cfg.maxIterations})

        # Collect results
        x = results['x']
        hess = Hessian(ModelLogistic.objectiveFunction,method='forward')(x,sumYBVTT, BVTT, y_regress)
        se = np.sqrt(np.diag(np.linalg.inv(hess)))
        ll = -results['fun']
        exitflag = results['status']

        # Compute VTT
        vtt = x[1] + x[2]*((self.arrays.T-1)/self.arrays.T)*np.sum(self.arrays.Choice*self.arrays.BVTT,1)

        return x, se, vtt, init_ll ,ll, exitflag

    @staticmethod
    def objectiveFunction(x: np.ndarray, sumYBVTT: np.ndarray, BVTT: np.ndarray, y_regress: np.ndarray):
        
        # Separate parameters: x is the estimated (multi-dimensional) parameter
        scale, intercept, parameter = x

        # Create value functions
        VTT = intercept + parameter * sumYBVTT
        V1 = scale * BVTT
        V2 = scale * VTT

        dV = V1 - V2
        dV[dV>700] = 700

        # Create choice probability and Log-likelihood
        p = 1 / (1 + np.exp(-dV))
        ll = np.log(p * (y_regress == 0) + (1 - p) * (y_regress == 1))
        ll[~np.isfinite(ll)] = 0
        
        # Return choice probability
        return -np.sum(ll)


#  Copyright 2021 Technische Universiteit Delft
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""Methods to create model arrays and generate descriptive statistics."""

import math
from typing import List

import pandas as pd
import numpy as np

from py_np4vtt.data_format import Vars, VarsMapping, DescriptiveStatsBasic, ModelArrays, Arrays


class VarMappingException(Exception):
    def __init__(self, missingVar: Vars, colName: str):
        self.missingVar = missingVar
        self.colName = colName

    def __str__(self):
        return f"The variable '{self.missingVar}' (mapped to column '{self.colName}') is missing from the dataset"


def make_arrays(dataset_frame: pd.DataFrame, dataset_varmapping: VarsMapping) -> Arrays:
    studied_arrays = {}

    for v in Vars:
        colName = dataset_varmapping[v]
        arr = dataset_frame.get(colName)
        if arr is None:
            raise VarMappingException(v, colName)
        else:
            studied_arrays[v] = arr

    return studied_arrays


def validate_modeldata(id_all, t, cost1, cost2, time1, time2, slow_alt, cheap_alt, choice) -> List[str]:
    # Create errormessage list
    errorList = []

    if not np.isfinite(id_all).all():
        errorList.append('There are either NAs or (minus) infinite values in ID Variable')

    if not (int(t) == t):
        errorList.append('Number of choice situations must be equal for all individuals.')

    if not np.isfinite(cost1).all():
        errorList.append('There are either NAs or (minus) infinite values in Cost of alternative 1.')

    if not np.isfinite(cost2).all():
        errorList.append('There are either NAs or (minus) infinite values in Cost of alternative 2.')

    if not np.isfinite(time1).all():
        errorList.append('There are either NAs or (minus) infinite values in Time of alternative 1.')

    if not np.isfinite(time2).all():
        errorList.append('There are either NAs or (minus) infinite values in Time of alternative 2.')

    if not (cheap_alt == slow_alt).all():
        errorList.append('At least one choice situation have either a cheap-fast or expensive-slow alternative.')

    if not np.logical_or((choice == 1), (choice == 2)).all():
        errorList.append('Chosen alternative variable must be either 1 or 2.')

    # Whoever calls this validator knows that empty errorList means validator success
    return errorList


def make_modelarrays(dataset_frame: pd.DataFrame, dataset_varmapping: VarsMapping) -> ModelArrays:
    """Create model arrays.

    This function takes a Pandas `DataFrame` and a dictionary that contains 
    the mapping between the necessary variables and the variable names in 
    the dataset.

    Parameters
    ----------
    dataset_frame : pandas.DataFrame
        A Pandas `DataFrame` that contains the dataset.

    dataset_varmapping: Dict[Vars,str]
        A dictionary file that maps the necessary variables required by NP4VTT 
        with the variable names in the dataset. Each key is of the format 
        `Vars.variablename` where `variablename` is one of the required 
        variables (see the documentation of `py_np4vtt.data_format.Vars` for 
        more details about the required variables). Each value is a string 
        that contains the name of the variable corresponding to the necessary 
        variable as it appears in `dataset_frame`.

    Returns
    -------
    ModelArrays
        An object containing the model arrays used by the nonparametric models.
    """
    study_arrays = make_arrays(dataset_frame, dataset_varmapping)

    # Copy to avoid changing the original data imported
    cost1 = study_arrays[Vars.Cost1].to_numpy(copy=True)
    cost2 = study_arrays[Vars.Cost2].to_numpy(copy=True)
    time1 = study_arrays[Vars.Time1].to_numpy(copy=True)
    time2 = study_arrays[Vars.Time2].to_numpy(copy=True)
    choice = study_arrays[Vars.ChosenAlt].to_numpy(copy=True)

    # Identify times and costs
    t1 = np.c_[time1,time2].max(axis=1)   # Higher time
    t2 = np.c_[time1,time2].min(axis=1)   # Lower time
    c1 = np.c_[cost1,cost2].min(axis=1)   # Lower cost
    c2 = np.c_[cost1,cost2].max(axis=1)   # Higher cost


    # Identify expensive and slow alternative
    cheap_alt = np.c_[cost1,cost2].argmin(axis=1) + 1
    slow_alt = np.c_[time1,time2].argmax(axis=1) + 1

    # Create scalars and ID variables
    id_all = study_arrays[Vars.Id]
    id_uniq = pd.unique(id_all)
    npar = id_uniq.size
    t = id_all.size / id_uniq.size

    errorList = validate_modeldata(id_all, t, c1, c2, t1, t2, slow_alt, cheap_alt, choice)
    t_int = math.floor(t)

    # BVTT
    bvtt = (- (c1-c2)/(t1-t2)).reshape((npar,t_int))

    # FBE = "Fast But Expensive"
    fbe_chosen = (choice != cheap_alt).reshape((npar, t_int))

    # The number of times a DM accepted the 'FBE' alt. Sum accross columns
    accepts = np.sum(fbe_chosen.astype(int), 1)

    return ModelArrays(
        BVTT=bvtt,
        Choice=fbe_chosen,
        Accepts=accepts,
        ID=id_uniq,
        NP=npar,
        T=t_int,
    )


def compute_descriptives(arrs: ModelArrays) -> DescriptiveStatsBasic:
    """Compute descriptive statistics
    
    It takes the object that contains the model arrays and returns a set of 
    relevant descriptive statistics regarding respondents, their choices and 
    about the BVTT.
    
    Parameters
    ----------
    arrs : ModelArrays
        The model arrays object

    Returns
    -------

    DescriptiveStatsBasic
        An object that contains the relevant descriptive statistics. Can be 
        accessed using `print()`.
    """
    fbe_units = arrs.Choice.astype(int)

    chosenBVTT = (fbe_units * arrs.BVTT)

    # noinspection PyTypeChecker
    chosenBVTT_mean: int = np.sum(chosenBVTT)/np.sum(fbe_units)

    chosen_fastexp = np.sum(fbe_units, 1)
    nt_cheapslow = np.count_nonzero(chosen_fastexp == 0)
    nt_fastext = np.count_nonzero(chosen_fastexp == arrs.T)

    return DescriptiveStatsBasic(
        NP=arrs.NP,
        T=arrs.T,
        NT_FastExp=nt_fastext,
        NT_CheapSlow=nt_cheapslow,
        ChosenBVTT_Mean=np.round(chosenBVTT_mean,4),
        BVTT_min=np.round(np.amin(arrs.BVTT),4),
        BVTT_max=np.round(np.amax(arrs.BVTT),4),
    )



#  Copyright 2021 Technische Universiteit Delft
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""Data format classes"""

from enum import Enum, auto
from dataclasses import dataclass
from typing import Dict
from inspect import cleandoc

import pandas as pd
import numpy as np
import numpy.typing as npt


class Vars(Enum):
    Id = auto()
    ChosenAlt = auto()
    Cost1 = auto()
    Time1 = auto()
    Cost2 = auto()
    Time2 = auto()


VarsMapping = Dict[Vars, str]

Arrays = Dict[Vars, pd.Series]


@dataclass
class ModelArrays:
    BVTT: npt.NDArray[np.float64]  # pd.Series[float]
    Choice: npt.NDArray[np.bool_]
    Accepts: npt.NDArray[np.int_]
    ID: npt.NDArray[np.int_]  # Unique participant IDs
    NP: int  # Number of participants
    T: int  # Number of choice situations per participant


@dataclass
class DescriptiveStatsBasic:
    NP: int  # Number of participants
    T: int  # Number of choice situations per participant
    NT_FastExp: int  # Number of non-traders (fast-expensive alt.)
    NT_CheapSlow: int  # Number of non-traders (cheap-slow alt.)
    ChosenBVTT_Mean: float  # Mean chosen BVTT
    BVTT_min: float  # Minimum of BVTT
    BVTT_max: float  # Maximum of BVTT

    def __str__(self) -> str:
        formatted_descriptives = cleandoc(f"""
            No. individuals: {self.NP}
            Sets per indiv.: {self.T}

            Number of non-traders:
            Fast-exp. alt.: {self.NT_FastExp}
            Slow-cheap alt.: {self.NT_CheapSlow}

            BVTT statistics:
            Mean chosen BVTT: {self.ChosenBVTT_Mean}
            Minimum of BVTT: {self.BVTT_min}
            Maximum of BVTT: {self.BVTT_max}
        """)

        return formatted_descriptives



#  Copyright 2021 Technische Universiteit Delft
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""Modules to configure and estimate a Local logit model."""
from dataclasses import dataclass
import numpy as np
from scipy.optimize import minimize
from py_np4vtt.data_format import ModelArrays
from py_np4vtt.utils import vtt_midpoints, predicted_vtt

@dataclass
class ConfigLocLogit:
    """Configuration class of the local logit model.
    
    This class stores the configuration parameters of a local logit model 
    and performs integrity checks before being passed to the model object.
    
    Parameters
    ----------
    
    minimum : float
        Minimum value of the VTT grid
    maximum : float
        Maximum value of the VTT grid
    supportPoints : int
        Number of support points of the VTT grid. The VTT grid will contain
        `(supportPoints-1)` intervals. Must be greater than zero
    """
    minimum: float
    maximum: float
    supportPoints: int

    def validate(self):
        # Create errormessage list
        errorList = []

        if not self.maximum > self.minimum:
            errorList.append('Max must be greater than minimum.')

        if not self.supportPoints > 0:
            errorList.append('No. of support points must be greater than zero.')

        # Whoever calls this validator knows that empty errorList means validator success
        return errorList

class ModelLocLogit:
    """Local logit model.
    
    This is the model class that prepares the data and estimates a local 
    logit model [1].
    
    Parameters
    -----------
    params : ConfigLocLogit
        A configuration class of a local logit model.
    arrays : ModelArrays
        Model arrays created with `make_modelarrays`
        
    Attributes
    ----------
    vtt_grid : numpy.ndarray
        The VTT grid created with the specifications of `ConfigLocLogit`.
    vtt_mid : numpy.ndarray
        The mid points of the VTT grid.

    Methods
    -------
    run():
        Estimates the local logit model.
    
    References
    ----------
    [1] Fosgerau, Mogens. "Using nonparametrics to specify a model to measure 
    the value of travel time." Transportation Research Part A: Policy and 
    Practice 41.9 (2007): 842-856.
    """
    def __init__(self, params: ConfigLocLogit, arrays: ModelArrays):
        self.params = params
        self.arrays = arrays

        # Create grid of support points
        self.vtt_grid = np.linspace(self.params.minimum, self.params.maximum, self.params.supportPoints)

        # Compute the midpoints of the VTT grid
        self.vtt_mid = vtt_midpoints(self.vtt_grid)

        # Compute distance between points at each support point
        dist = self.vtt_grid[1] - self.vtt_grid[0]

        # Print message of the support points
        print("Created a VTT grid of " + str(self.params.supportPoints) + \
            " points between " + str(self.params.minimum) + " and " + str(self.params.maximum) + ".")

        print("Distance between points of the VTT grid is " + str(dist))
        
    def run(self):
        """Estimates the local logit model.
        
        Parameters
        ----------
        None.

        Returns
        -------
        p : numpy.ndarray
            The estimates of the cumulative choice probability (CDF), evaluated 
            at each interval of the VTT grid. The first point is always zero 
            while the last point is always equal to the second last for 
            compatibility with plots.
        vtt : numpy.ndarray
            The estimated VTT per respondent, based in the estimated 
            CDF points (`p`) and the sample.
        ll : float
            The log-likelihood function at the optimum of the estimation.
        """
        # Compute the kernel width
        k = np.r_[self.vtt_grid, 0.] - np.r_[0., self.vtt_grid]
        k = k[:-2]
        k[0] = k[1].copy()

        YX = self.arrays.Choice.T.flatten()

        # Perform a weighted logit for each support point
        p = []
        fval = 0.
        for n in range(len(self.vtt_grid)-1):
            x, fval_x = ModelLocLogit.initLocalLogit(n, k[n], self.arrays.BVTT, YX, self.vtt_grid)
            p.append(x[0])
            fval = fval + fval_x
        
        # Return probability array and -ll
        p = np.array(p)
        ll = -fval

        # Compute the predicted VTT at the midpoints
        vtt = predicted_vtt(p,self.vtt_mid,self.arrays.NP)

        # Add point 0 in the estimated CDF and repeat last point to make coincide with point zero and last point in the VTT mid point
        p = np.concatenate((0,p,p[-1]),axis=None)

        return p, vtt, ll

    @staticmethod
    def initLocalLogit(n, k, BVTT, YX, vtt_grid):
        # Get observations
        BVTT_flat = BVTT.T.flatten()
        xn = BVTT_flat[(BVTT_flat > (vtt_grid[n]-k)) & (BVTT_flat < (vtt_grid[n] + k))]
        x0 = vtt_grid[n]
        y_local = YX[(BVTT_flat > (vtt_grid[n]-k)) & (BVTT_flat < (vtt_grid[n] + k))]
        dist = np.abs(x0-xn)
        weight = (k-dist)/k

        # Search function
        coef_start = np.array([0., 0.])
        args = (y_local, xn, x0, weight)
        results = minimize(ModelLocLogit.objectiveFunction, coef_start, args=args, method='L-BFGS-B',options={'gtol': 1e-6})

        # Collect results
        x = results['x']
        fval = results['fun']

        # Convert to probabilities
        x = 1 - (np.exp(x)/(1+np.exp(x)))

        return x, fval

    @staticmethod
    def objectiveFunction(coef: np.ndarray, y_local: np.ndarray, xn: np.ndarray, x0: np.ndarray, weight: np.ndarray):
        acc = coef[0] + coef[1]*(xn-x0)
        P = np.exp(acc)/(1+np.exp(acc))
        LL = np.log(P*(y_local == 1) + (1-P)*(y_local == 0))
        LLw = -sum(weight*LL)

        return LLw



#  Copyright 2021 Technische Universiteit Delft
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""Modules to configure and estimate an ANN-based VTT model."""
from dataclasses import dataclass

from typing import List, Optional

from numpy import ndarray

from py_np4vtt.data_format import ModelArrays

import numpy as np

from sklearn.neural_network import MLPClassifier
from sklearn.model_selection import train_test_split
from sklearn.metrics import log_loss

# TODO: Check with Joao how safe is to import os

@dataclass
class ConfigANN:
    """Configuration class of the ANN-based model.
    
    This class stores the configuration parameters of an ANN-based model 
    and performs integrity checks before being passed to the model object.
    
    Parameters
    ----------
    hiddenLayerNodes : List[int]
        Topology of the ANN. The length of `hiddenLayerNodes` is the number 
        of hidden layers, and each element of the list is the number of 
        nodes of the corresponding layer.
    trainingRepeats : int
        Number of estimation repeats of the ANN.
    shufflesPerRepeat : int
        Random shuffles of the data per repetition.

    seed : Optional[int]
        Random seed for the shuffling and estimation process.
    """
    hiddenLayerNodes: List[int]

    trainingRepeats: int
    shufflesPerRepeat: int

    seed: Optional[int]

    def validate(self):
        # Create errormessage list
        errorList = []

        if not self.trainingRepeats > 0:
            errorList.append('Number of repeats must be positive.')

        if not self.shufflesPerRepeat > 0:
            errorList.append('Number of shuffles per repeats must be positive.')

        # Whoever calls this validator knows that empty errorList means validator success
        return errorList

class ModelANN:
    """ANN model.
    
    This is the model class that prepares the data and estimates 
    the ANN-based model [1].
    
    Parameters
    -----------
    params : ConfigANN
        A configuration class of an ANN-based model.
    arrays : ModelArrays
        Model arrays created with `make_modelarrays`
        
    Attributes
    ----------
    X_train : numpy.ndarray
        Input data for training (estimation), i.e. the BVTT of each choice set and the T-1'th choices
    X_test : numpy.ndarray
        Input data for testing (prediction)
    X_full : numpy.ndarray
        Full input data
    y_train : numpy.ndarray
        Output data for training, i.e. the T'th choice
    y_test : numpy.ndarray
        Output data for testing
    y_full : numpy.ndarray
        Full output data

    Methods
    -------
    run():
        Estimates the ANN-based model.
    
    References
    ----------
    [1] van Cranenburgh, Sander, and Marco Kouwenhoven. "An artificial neural 
    network based method to uncover the value-of-travel-time distribution." 
    Transportation 48.5 (2021): 2545-2583.
    """
    def __init__(self, cfg: ConfigANN, arrays: ModelArrays):
        self.cfg = cfg
        self.arrays = arrays

        # Initialise arrays for randomisation
        shuffle_index = np.zeros((self.cfg.shufflesPerRepeat,self.arrays.T+1))
        full_data_array = np.zeros((self.arrays.NP,self.cfg.shufflesPerRepeat*(self.arrays.T+1),2))

        # Randomise data
        for n in range(self.arrays.NP):
            for m in range(self.cfg.shufflesPerRepeat):
                rnd11 = np.arange(self.arrays.T)
                np.random.shuffle(rnd11)
                shuffle_index[m,:] = np.hstack((rnd11,rnd11[np.random.randint(1,self.arrays.T-1)]))
            
            full_data_array[n,:,0] = self.arrays.Choice[n,shuffle_index.astype(int).flatten()]
            full_data_array[n,:,1] = self.arrays.BVTT[n,shuffle_index.astype(int).flatten()]

        # Create input and output arrays for ANN
        full_data_array = np.hstack((np.reshape(full_data_array[:,:,0].T,(self.arrays.T+1,self.cfg.shufflesPerRepeat*self.arrays.NP),order='F').T,np.reshape(full_data_array[:,:,1].T,(self.arrays.T+1,self.cfg.shufflesPerRepeat*self.arrays.NP),order='F').T))
        t = full_data_array[:,0]
        x = full_data_array[:,1:]

        # Separate in train and test
        X_train, X_test, y_train, y_test = train_test_split(x,t,test_size = 0.15,random_state=self.cfg.seed)

        self.X_train = X_train
        self.X_test = X_test
        self.X_full = x
        self.y_train = y_train
        self.y_test = y_test
        self.y_full = t

    def run(self, verbose=True):
        """Estimates the Rouwendal model.
        
        Parameters
        ----------
        None.

        Returns
        -------
        ll_list : np.ndarray
            The log-likelihood values per repetition.
        r2_list : np.ndarray
            The Rho-squared values per repetition.
        vtt_list : np.ndarray
            The estimated VTT per respondent and repetition.
        """
        ll_list = []
        rho_sq = []
        y_predict = []
        VTT_mid_list = []

        for r in range(self.cfg.trainingRepeats):
            
            if verbose:
                print('Rep ' + str(r+1) + ': ',end='',flush=True)

            clf = MLPClassifier(
                hidden_layer_sizes=self.cfg.hiddenLayerNodes,
                activation='tanh',
                tol=1e-4,
                alpha=0.,
                n_iter_no_change=6,
                max_iter=1000,
                early_stopping=True,
                validation_fraction=0.1275,
                verbose=False,random_state=self.cfg.seed).fit(self.X_train,self.y_train)

            # Predict in test sample
            y_predict_train = clf.predict_proba(self.X_train)
            y_predict_test = clf.predict_proba(self.X_test)

            y_predict.append(y_predict_test)

            # Get train and test loss
            train_loss = log_loss(self.y_train,y_predict_train)
            test_loss = log_loss(self.y_test,y_predict_test)

            # Compute log-likelihood and Rho-sq in full sample
            ll = -len(self.y_train)*train_loss - len(self.y_test)*test_loss
            r2 = 1 - (ll/(np.log(0.5)*len(self.y_full)))
            ll_list.append(ll)
            rho_sq.append(r2)

            if verbose:
                print('CE (train): ' + str(round(train_loss,4)) + ' / CE (test): ' + str(round(test_loss,4)) + ' / LL: ' + str(round(ll,2)) + ' / Rho-sq: ' + str(round(r2,2)))
            
            # Simulate N-choice of each individual using the ANN
            vtt_grid = np.tile(np.linspace(0,1.5*self.X_full.max(),101),(self.arrays.Accepts.shape[0],1))
            y_pred_N = ModelANN.simulateNChoice(self,clf,self.arrays.Choice,vtt_grid,self.arrays.BVTT,20)

            # Recover individual VTTs, using simulation of choice probs
            VTT_mid = np.zeros(self.arrays.NP)
            for n in range(self.arrays.NP):
                if np.max(y_pred_N[n,:]) > 0.5 and np.min(y_pred_N[n,:])<0.5:
                    delta_x = vtt_grid[n,np.where((y_pred_N[n,:]-0.5)>=0)[0][-1]] - vtt_grid[n,np.where((y_pred_N[n,:]-0.5)<=0)[0][0]]
                    delta_y = y_pred_N[n,np.where((y_pred_N[n,:]-0.5)>=0)[0][-1]] - y_pred_N[n,np.where((y_pred_N[n,:]-0.5)<=0)[0][0]]
                    VTT_mid[n] = vtt_grid[n,np.where((y_pred_N[n,:]-0.5)>=0)[0][-1]] - (y_pred_N[n,np.where((y_pred_N[n,:]-0.5)>=0)[0][-1]]-0.5)/(delta_y/delta_x)

            VTT_mid_list.append(VTT_mid)

        ll_list = np.array(ll_list)
        r2_list = np.array(rho_sq)
        vtt_list = np.array(VTT_mid_list)

        return ll_list, r2_list, vtt_list

    @staticmethod
    def simulateNChoice(self,clf,y,vtt_grid,X,R):

        # Create arrays for random shufflings
        x_sim = np.zeros((y.shape[0]*vtt_grid.shape[1],2*self.arrays.T+1,R))
        y_sim = np.zeros((y.shape[0],vtt_grid.shape[1],R))
        
        for n in range(y.shape[0]):
            for r in range(R):
                rndp = np.arange(self.arrays.T)
                np.random.shuffle(rndp)

                x_sim[(n*vtt_grid.shape[1]):((n+1)*vtt_grid.shape[1]),:,r] = np.c_[np.tile(y[n,rndp],(vtt_grid.shape[1],1)), vtt_grid[n,:].T, np.tile(X[n,rndp],(vtt_grid.shape[1],1))]
        
        for r in range(R):
            y_sim[:,:,r] = np.reshape(clf.predict_proba(x_sim[:,:,r])[:,1],(vtt_grid.shape[1],y.shape[0]),order='F').T
        y_median = np.median(y_sim,axis=2)

        return y_median